use library_loader::load_function_ptrs_from_lib;

load_function_ptrs_from_lib!(VitureLibrary, "libs/libviture_one_sdk.so", {
    init(
        imu: extern fn(*const u8, u16, u32),
        mcu: extern fn(u16, *const u8, u16, u32)
    ) -> bool,
    deinit() -> (),
    set_imu(on_off: bool) -> i32,
    get_imu_state() -> i32,
    set_3d(on_off: bool) -> i32,
    get_3d_state() -> i32,
});

static mut IMU_CALLBACK: Option<Box<dyn Fn(&[u8], u32) -> ()>> = None;
static mut MCU_CALLBACK: Option<Box<dyn Fn(u16, &[u8], u32) -> ()>> = None;

extern "C" fn imu(data: *const u8, len: u16, ts: u32) {
    unsafe {
        if let Some(imu_callback) = &IMU_CALLBACK {
            imu_callback(std::slice::from_raw_parts(data, len as usize), ts);
        }
    }
}

extern "C" fn mcu(msgid: u16, data: *const u8, len: u16, ts: u32) {
    unsafe {
        if let Some(mcu_callback) = &MCU_CALLBACK {
            mcu_callback(msgid, std::slice::from_raw_parts(data, len as usize), ts);
        }
    }
}

#[allow(non_camel_case_types)]
#[repr(C)]
#[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
pub enum VitureResult {
    ERR_TIMEOUT = -3,
    ERR_RSP_ERROR = -2,
    ERR_WRITE_FAIL = -1,

    ERR_SUCCESS = 0,
    ERR_FAILURE = 1,
    ERR_INVALID_ARGUMENT = 2,
    ERR_NOT_ENOUGH_MEMORY = 3,
    ERR_UNSUPPORTED_CMD = 4,
    ERR_CRC_MISMATCH = 5,
    ERR_VER_MISMATCH = 6,
    ERR_MSG_ID_MISMATCH = 7,
    ERR_MSG_STX_MISMATCH = 8,
    ERR_CODE_NOT_WRITTEN = 9,
}

impl From<i32> for VitureResult {
    fn from(value: i32) -> Self {
        match value {
            -3 => Self::ERR_TIMEOUT,
            -2 => Self::ERR_RSP_ERROR,
            -1 => Self::ERR_WRITE_FAIL,

            0 => Self::ERR_SUCCESS,
            1 => Self::ERR_FAILURE,
            2 => Self::ERR_INVALID_ARGUMENT,
            3 => Self::ERR_NOT_ENOUGH_MEMORY,
            4 => Self::ERR_UNSUPPORTED_CMD,
            5 => Self::ERR_CRC_MISMATCH,
            6 => Self::ERR_VER_MISMATCH,
            7 => Self::ERR_MSG_ID_MISMATCH,
            8 => Self::ERR_MSG_STX_MISMATCH,
            9 => Self::ERR_CODE_NOT_WRITTEN,

            _ => panic!("unsupported value: {value}"),
        }
    }
}

pub struct Viture {
    library: VitureLibrary,
}

impl Viture {
    pub fn new<IMU, MCU>(imu_callback: IMU, mcu_callback: MCU) -> Result<Self, VitureResult>
    where
        IMU: Fn(&[u8], u32) -> () + 'static,
        MCU: Fn(u16, &[u8], u32) -> () + 'static,
    {
        let library = VitureLibrary::load().map_err(|err| {
            println!("Err: {err:?}");
            VitureResult::ERR_FAILURE
        })?;

        unsafe {
            IMU_CALLBACK = Some(Box::new(imu_callback));
            MCU_CALLBACK = Some(Box::new(mcu_callback));
        }

        let me = Self { library };

        if !unsafe { me.library.init(imu, mcu) } {
            return Err(VitureResult::ERR_FAILURE);
        }

        Ok(me)
    }

    pub fn set_imu(&self, on_off: bool) -> Result<(), VitureResult> {
        unsafe {
            let result = VitureResult::from(self.library.set_imu(on_off));

            if result == VitureResult::ERR_SUCCESS {
                Ok(())
            } else {
                Err(result)
            }
        }
    }

    pub fn get_imu(&self) -> Result<bool, VitureResult> {
        unsafe {
            let result = self.library.get_imu_state();

            match result {
                0 => Ok(false),
                1 => Ok(true),
                _ => Err(VitureResult::from(result)),
            }
        }
    }

    pub fn set_3d(&self, on_off: bool) -> Result<(), VitureResult> {
        unsafe {
            let result = VitureResult::from(self.library.set_3d(on_off));

            if result == VitureResult::ERR_SUCCESS {
                Ok(())
            } else {
                Err(result)
            }
        }
    }

    pub fn get_3d(&self) -> Result<bool, VitureResult> {
        unsafe {
            let result = self.library.get_3d_state();

            match result {
                0 => Ok(false),
                1 => Ok(true),
                _ => Err(VitureResult::from(result)),
            }
        }
    }
}

impl Drop for Viture {
    fn drop(&mut self) {
        unsafe {
            self.library.deinit();
        }
    }
}

#[cfg(test)]
mod test {
    #[test]
    fn initialisation() -> anyhow::Result<()> {
        let _viture = super::Viture::new(|_, _| {}, |_, _, _| {})
            .map_err(|err| anyhow::anyhow!("{err:?}"))?;

        Ok(())
    }
}
